import dataclasses
import json
import logging
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Parse json string to dictionary.

    If parsing fails, return a dictionary, the value of which is either provided in the Config or {'failed_parse': raw_str} by default.

    Inputs:
        json_str (str): JSON string.

    Config:
        default_value (dict): In the event of a failed parse, this value is used as the parsed value. If not provided, a default dictionary {'failed_parse': raw_str} is used.

    Outputs:
        result (dict): Parsed dictionary from the input JSON string, or default_value dict when the input cannot be parsed as JSON.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        json_str: str

    @dataclasses.dataclass
    class Config:
        default_value: dict | None = None

    @dataclasses.dataclass
    class Outputs:
        result: dict

    def execute(self, inputs):
        # Remove markdown JSON block if present
        json_str = inputs.json_str.replace('```json', '').replace('```', '') if '```json' in inputs.json_str else inputs.json_str
        try:
            result = json.loads(json_str)
        except Exception as e:
            logger.info(f"Could not decode JSON string: {e}")
            result = self.config.default_value if self.config.default_value is not None else {'failed_parse': inputs.json_str}
        return self.Outputs(result=result)

    def dry_run(self, inputs):
        return self.execute(inputs)



import copy
import json
import unittest
from jinja2 import Template

from irisml.tasks.get_system_prompt_from_key_value_pair_schema import Task


class TestGetSystemPromptFromKeyValuePairSchema(unittest.TestCase):
    def get_schema(self):
        return copy.deepcopy({
            "name": "UI automation Schema",
            "description": "Find UI elements and actions by natural language query from given image.",
            "fieldSchema": {
                "UIElementBox": {
                    "type": "bbox",
                    "description": "bounding box coordinates of ui element in [left, top, width, height] absolute pixel values."
                }
            }
        })

    def test_default_template(self):
        schema = self.get_schema()
        inputs = Task.Inputs(schema)
        outputs = Task(Task.Config()).execute(inputs)

        prompt = Template(Task.TEMPLATE).render(schema_name=schema['name'], schema_description=schema['description'], field_schema=json.dumps(schema['fieldSchema'], indent=2))
        self.assertEqual(outputs.prompt, prompt)

    def test_default_template_no_field_schema(self):
        schema = self.get_schema()
        del schema['fieldSchema']
        inputs = Task.Inputs(schema)
        outputs = Task(Task.Config()).execute(inputs)

        prompt = Template(Task.TEMPLATE).render(schema_name=schema['name'], schema_description=schema['description'], field_schema=None)
        self.assertEqual(outputs.prompt, prompt)

    def test_custom_template(self):
        schema = self.get_schema()
        inputs = Task.Inputs(schema)

        template = '''Write response according to schema: {{schema_name}}. {{schema_description}}\n{{field_schema}}'''
        outputs = Task(Task.Config(jinja_template=template)).execute(inputs)
        self.assertEqual(outputs.prompt, '''Write response according to schema: UI automation Schema. Find UI elements and actions by natural language query from given image.
{
  "UIElementBox": {
    "type": "bbox",
    "description": "bounding box coordinates of ui element in [left, top, width, height] absolute pixel values."
  }
}''')



import unittest
from irisml.tasks.append_to_str_list import Task


class TestAppendToStrList(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs(str_list=['a', 'b', 'c'], item='d'))
        self.assertEqual(outputs.str_list, ['a', 'b', 'c', 'd'])



import dataclasses
import logging
import pathlib
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Reads a file and returns its contents as string.

    Config:
        path (pathlib.Path): Path to the file to read.
    """
    VERSION = '0.1.1'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path

    @dataclasses.dataclass
    class Outputs:
        data: str

    def execute(self, inputs):
        if not self.config.path.exists():
            raise ValueError(f"File not found: {self.config.path}")

        data = pathlib.Path(self.config.path).read_text()
        logger.info(f"Read str from {self.config.path}")
        return self.Outputs(data)

    def dry_run(self, inputs):
        return self.execute(inputs)



from collections import defaultdict
import dataclasses
import logging
import typing
import torch

import irisml.core

logger = logging.getLogger(__name__)


# predictions -> topk indices -> get dataset ids from this task -> ICL task

class Task(irisml.core.TaskBase):
    """Go through a flat array to get indices of all items matching given values. Return a list[list[int]] in case of multiple matches.
    Support values as flat list or nested list (2-level). For nested list, concatenate all matched indices for each item in the nested list.
    flat list: e.g. array = [1, 2, 3, 3], values = [1, 3], return [[0], [2, 3]]
    nested list: e.g. array = [1, 2, 3, 3], values = [[1, 2], [3, 4], [5]], return [[0, 1], [2, 3], []].

    Inputs:
        array_list (list): List of items to search through
        array_tensor (torch.Tensor): Tensor of items to search through, must provide array_list or array_tensor
        values_list (list): List of items to search for
        values_tensor (torch.Tensor): Tensor of items to search for, must provide values_list or values_tensor
    """
    VERSION = '0.0.1'

    @dataclasses.dataclass
    class Inputs:
        array_list: typing.Optional[list] = None
        array_tensor: typing.Optional[torch.Tensor] = None
        values_list: typing.Optional[list] = None
        values_tensor: typing.Optional[torch.Tensor] = None

    @dataclasses.dataclass
    class Outputs:
        indices: list[list[int]]

    def execute(self, inputs):
        if inputs.array_tensor is not None:
            array = inputs.array_tensor.tolist()
        elif inputs.array_list is not None:
            array = inputs.array_list
        else:
            raise ValueError('Array must be provided!')
        if any(isinstance(a, list) for a in array):
            raise ValueError('Support only flat array!')

        if inputs.values_tensor is not None:
            values = inputs.values_tensor.tolist()
        elif inputs.values_list is not None:
            values = inputs.values_list
        else:
            raise ValueError('Values must be provided!')

        # Convert values to 2d list and check it is 2d
        values = [v if isinstance(v, list) else [v] for v in values]
        if any(isinstance(v2, list) for v in values for v2 in v):
            raise ValueError('Support at most 2-level nested list for values!')

        idx_by_val = defaultdict(list)
        for i, val in enumerate(array):
            idx_by_val[val].append(i)

        indices = []
        for val in values:
            val_indices = []
            for v in val:
                val_indices.extend(idx_by_val.get(v, []))
            indices.append(val_indices)

        return Task.Outputs(indices)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
from irisml.tasks.decode_json_str import Task


class TestDecodeJsonStr(unittest.TestCase):
    def test_simple(self):
        inputs = Task.Inputs('{"a": 1}')
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.result, {'a': 1})

        inputs = Task.Inputs('{{"b": 2}')
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.result, {'failed_parse': '{{"b": 2}'})

    def test_remove_json_markdown_block(self):
        inputs = Task.Inputs('```json\n{"a": 1}   \n```')
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.result, {'a': 1})

        inputs = Task.Inputs('```json{"a": 1}```')
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.result, {'a': 1})

    def test_simple_specify_default_value(self):
        inputs = Task.Inputs('{"a": 1}')
        outputs = Task(Task.Config(default_value={"c": None})).execute(inputs)
        self.assertEqual(outputs.result, {'a': 1})

        inputs = Task.Inputs('{{"b": 2}')
        outputs = Task(Task.Config(default_value={"c": None})).execute(inputs)
        self.assertEqual(outputs.result, {'c': None})



import dataclasses
import json
import logging
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Parse a list of json strings to dictionaries.

    If parsing fails, return a dictionary, the value of which is either provided in the Config or {'failed_parse': raw_str} by default.

    Inputs:
        json_strs (List[str]): JSON strings.

    Config:
        default_value (dict): In the event of a failed parse, this value is used as the parsed value. If not provided, a default dictionary {'failed_parse': raw_str} is used.

    Outputs:
        dicts (List[dict]): Parsed dictionaries from the input JSON strings, or default_value dicts when the input cannot be parsed as JSON.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        json_strs: list[str]

    @dataclasses.dataclass
    class Config:
        default_value: dict | None = None

    @dataclasses.dataclass
    class Outputs:
        dicts: list[dict]

    def execute(self, inputs):
        dicts = []
        for json_str in inputs.json_strs:
            # Remove markdown json block if present
            json_str = json_str.replace('```json', '').replace('```', '') if '```json' in json_str else json_str
            try:
                dicts.append(json.loads(json_str))
            except Exception as e:
                logger.info(f"Could not decode JSON string: {e}")
                dicts.append(self.config.default_value if self.config.default_value is not None else {'failed_parse': json_str})
        return self.Outputs(dicts=dicts)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.get_array_indices_from_values import Task


class TestGetArrayIndicesFromValues(unittest.TestCase):
    def test_flat(self):
        array = [1, 2, 3, 3]
        values = [1, 2, 5]
        outputs = Task(Task.Config()).execute(Task.Inputs(array_list=array, values_list=values))
        self.assertEqual(outputs.indices, [[0], [1], []])

    def test_nested(self):
        array = [1, 2, 3, 3]
        values = [[1, 2], [3, 4], [5]]
        outputs = Task(Task.Config()).execute(Task.Inputs(array_list=array, values_list=values))
        self.assertEqual(outputs.indices, [[0, 1], [2, 3], []])

    def test_tensor(self):
        array = [1, 2, 3, 3]
        values_tensor = torch.tensor([[1, 2], [3, 4]])
        outputs = Task(Task.Config()).execute(Task.Inputs(array_list=array, values_tensor=values_tensor))
        self.assertEqual(outputs.indices, [[0, 1], [2, 3]])

        outputs = Task(Task.Config()).execute(Task.Inputs(array_tensor=torch.tensor(array), values_tensor=values_tensor))
        self.assertEqual(outputs.indices, [[0, 1], [2, 3]])

    def test_empty_array(self):
        array = []
        values = [[1, 2], [3, 4]]
        outputs = Task(Task.Config()).execute(Task.Inputs(array_list=array, values_list=values))
        self.assertEqual(outputs.indices, [[], []])



# flake8: noqa: E501

import dataclasses
import json
import logging
from jinja2 import Template

import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Get a system prompt from the schema of a KeyValuePair dataset (https://github.com/microsoft/vision-datasets/blob/main/COCO_DATA_FORMAT.md#keyvaluepair-dataset).

    Inputs:
        schema (dict): The schema. An example for a VQA task is 
            {
                "name": "visual question answering schema",
                "description": "",
                "fieldSchema": {
                    "answer": {
                        "type": "string",
                        "description": "Answer to the question and image."
                    }
                }
            }
            More examples can be found at https://github.com/microsoft/vision-datasets/blob/main/DATA_PREPARATION.md and https://github.com/microsoft/vision-datasets/blob/main/tests/resources/util.py.

    Config:
        jinja_template: str = None. jinja template str for the system prompt which should contain three variables: schema_name, schema_description, field_schema. The will be filled with 'name',
            'description', and 'fieldSchema' from schema dictionary respectively. If jinja_template is not provided, a default template will be used.
    """
    VERSION = '0.1.1'

    TEMPLATE = '''Extract information from images in a JSON format following the provided Schema.
# Schema name{% if schema_description is not none %} and description{% endif %}
{{schema_name}}. {{schema_description}}
{% if field_schema is not none %}# JSON Schema definition
```json
{{field_schema}}
```{% endif %}
'''

    @dataclasses.dataclass
    class Inputs:
        schema: dict

    @dataclasses.dataclass
    class Config:
        jinja_template: str | None = None

    @dataclasses.dataclass
    class Outputs:
        prompt: str

    def execute(self, inputs):
        template = Template(self.config.jinja_template if self.config.jinja_template is not None else self.TEMPLATE)
        field_schema = json.dumps(inputs.schema['fieldSchema'], indent=2) if 'fieldSchema' in inputs.schema else None
        prompt = template.render(schema_name=inputs.schema['name'], schema_description=inputs.schema.get('description', ''), field_schema=field_schema)
        logger.info(f'Generated system prompt:\n{prompt}')

        return self.Outputs(prompt)

    def dry_run(self, inputs):
        return self.execute(inputs.schema)



import dataclasses
import irisml.core


class Task(irisml.core.TaskBase):
    """Append a string to a list of strings.

    Inputs:
        str_list (list[str]): List of strings.
        item (str): String to append to the list.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        str_list: list[str]
        item: str

    @dataclasses.dataclass
    class Outputs:
        str_list: list[str]

    def execute(self, inputs):
        if not isinstance(inputs.str_list, list):
            raise ValueError(f'Expected str_list to be a list, got {type(inputs.str_list)}')
        if not isinstance(inputs.item, str):
            raise ValueError(f'Expected item to be a str, got {type(inputs.item)}')

        return self.Outputs(str_list=inputs.str_list + [inputs.item])

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import json
import logging
import pathlib
import typing
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Save a list of tensors to a JSONL file.

    Config:
        path (Path): The path to save the JSONL file.
    """
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        data: typing.List[torch.Tensor]

    @dataclasses.dataclass
    class Config:
        path: pathlib.Path

    def execute(self, inputs):
        self.config.path.parent.mkdir(parents=True, exist_ok=True)

        with self.config.path.open('w') as f:
            for item in inputs.data:
                json.dump(item.tolist(), f)
                f.write('\n')

        logger.info(f'Saved {len(inputs.data)} tensors to {self.config.path}')
        return self.Outputs()



import dataclasses
import io
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Average a list of state_dicts.

    This task accepts a list of state dict objects or a list of bytes objects that is a serialized state_dict.

    All state_dicts must have the same keys.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        state_dict_list: list[dict] | None = None
        state_dict_bytes_list: list[bytes] | None = None

    @dataclasses.dataclass
    class Outputs:
        state_dict: dict

    def execute(self, inputs):
        num_sources = len([x for x in [inputs.state_dict_list, inputs.state_dict_bytes_list] if x])
        if num_sources != 1:
            raise ValueError("One state_dict source must be provided.")

        state_dict_list = None
        if inputs.state_dict_list:
            state_dict_list = inputs.state_dict_list
        if inputs.state_dict_bytes_list:
            state_dict_list = [torch.load(io.BytesIO(x), map_location='cpu') for x in inputs.state_dict_bytes_list]

        assert state_dict_list, "Failed to load the state_dict."

        logger.info(f"Loaded {len(state_dict_list)} state_dicts.")

        # Verify the key names
        key_names = [set(x.keys()) for x in state_dict_list]
        for i in range(1, len(key_names)):
            extra_keys = key_names[i] - key_names[0]
            if extra_keys:
                raise ValueError(f"State dict {i} has extra keys: {extra_keys}")
            missing_keys = key_names[0] - key_names[i]
            if missing_keys:
                raise ValueError(f"State dict {i} is missing keys: {missing_keys}")

        state_dict = {}
        for k in state_dict_list[0].keys():
            state_dict[k] = sum([x[k] for x in state_dict_list]) / len(state_dict_list)

        return self.Outputs(state_dict=state_dict)

    def dry_run(self, inputs):
        return self.execute(inputs)



import dataclasses
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Reshape a tensor using the given shape.

    Config:
        shape (list[int]): The new shape. See torch.Tensor.reshape() method for the detail.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        tensor: torch.Tensor

    @dataclasses.dataclass
    class Config:
        shape: list[int]

    @dataclasses.dataclass
    class Outputs:
        tensor: torch.Tensor

    def execute(self, inputs):
        tensor = inputs.tensor.reshape(*self.config.shape)
        logger.info(f"Reshaped {inputs.tensor.shape} into {tensor.shape}")
        return self.Outputs(tensor=tensor)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
from irisml.tasks.decode_json_strs import Task


class TestDecodeJsonStrs(unittest.TestCase):
    def test_simple(self):
        inputs = Task.Inputs(['{"a": 1}', '{"b": 2}'])
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.dicts[0], {'a': 1})
        self.assertEqual(outputs.dicts[1], {'b': 2})

        inputs = Task.Inputs(['{"a": 1}', '{{"b": 2}'])
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.dicts[0], {'a': 1})
        self.assertEqual(outputs.dicts[1], {'failed_parse': '{{"b": 2}'})

    def test_simple_specify_default_value(self):
        inputs = Task.Inputs(['{"a": 1}', '{"b": 2}'])
        outputs = Task(Task.Config(default_value={"c": None})).execute(inputs)
        self.assertEqual(outputs.dicts[0], {'a': 1})
        self.assertEqual(outputs.dicts[1], {'b': 2})

        inputs = Task.Inputs(['{"a": 1}', '{{"b": 2}'])
        outputs = Task(Task.Config(default_value={"c": None})).execute(inputs)
        self.assertEqual(outputs.dicts[0], {'a': 1})
        self.assertEqual(outputs.dicts[1], {'c': None})

    def test_remove_json_markdown_block(self):
        inputs = Task.Inputs(['```json\n{"a": 1}   \n```'])
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.dicts[0], {'a': 1})

        inputs = Task.Inputs(['```json{"a": 1}```'])
        outputs = Task(Task.Config()).execute(inputs)
        self.assertEqual(outputs.dicts[0], {'a': 1})



import dataclasses
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Normalize the last dimension of a tensor with L2 norm."""
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        tensor: torch.Tensor

    @dataclasses.dataclass
    class Outputs:
        tensor: torch.Tensor

    def execute(self, inputs):
        tensor = inputs.tensor / inputs.tensor.norm(dim=-1, keepdim=True)
        return self.Outputs(tensor=tensor)

    def dry_run(self, inputs):
        return self.execute(inputs)



import unittest
import torch
from irisml.tasks.reshape_tensor import Task


class TestReshapeTensor(unittest.TestCase):
    def test_simple(self):
        tensor = torch.rand((2, 3, 4))
        outputs = Task(Task.Config(shape=[2, 3, 2, 2])).execute(Task.Inputs(tensor))
        self.assertEqual(outputs.tensor.shape, (2, 3, 2, 2))
        self.assertTrue(torch.equal(outputs.tensor, tensor.reshape(2, 3, 2, 2)))

    def test_variable(self):
        tensor = torch.rand((2, 3, 4))
        outputs = Task(Task.Config(shape=[-1, 1, 4])).execute(Task.Inputs(tensor))
        self.assertEqual(outputs.tensor.shape, (6, 1, 4))
        self.assertTrue(torch.equal(outputs.tensor, tensor.reshape(6, 1, 4)))

    def test_failure(self):
        tensor = torch.rand((2, 3, 4))
        with self.assertRaises(RuntimeError):
            Task(Task.Config(shape=[3, 3, 4])).execute(Task.Inputs(tensor))
